#include "stacktool.h"

avr_insn **all_code = NULL;
bool *sp_consistent = NULL;
data_addr_t SRAM_TOP = NO_ADDR, init_stack_ptr = NO_ADDR;

const bool DEBUG_DECODE = false;

// #include "avr_opcode_macros.ci"

const char* opcode_to_string (enum AVR_OPCODE opcode)
{
  switch (opcode) {

#include "avr_opcode_to_str.ci"

  case AVR_SET_STACK_PTR:
    return "set_SP";

  default:
    return NULL;
  }
}

bool avr_is_call(enum AVR_OPCODE opcode)
{
  return (opcode == AVR_call ||
          opcode == AVR_rcall ||
          opcode == AVR_icall ||
          opcode == AVR_eicall);
}

bool avr_is_ret(enum AVR_OPCODE opcode)
{
  return (opcode == AVR_ret ||
          opcode == AVR_reti);
}

bool avr_is_jmp(enum AVR_OPCODE opcode)
{
  return (opcode == AVR_jmp ||
          opcode == AVR_rjmp);
}

static int unsigned_to_twos_comp (int bits, int k)
{
  if (k > (1<<(bits-1))-1) {
    k -= 1<<bits;
  }
  return k;
}

static void set_dec_dst_reg (struct decoder_info *dec, unsigned int r)
{
  dec->dst_reg_set = true;
  dec->dst_reg = (reg_t) r;
}

static void set_dec_src_reg (struct decoder_info *dec, unsigned int r)
{
  dec->src_reg_set = true;
  dec->src_reg = (reg_t) r;
}

static void set_dec_immed (struct decoder_info *dec, unsigned int i)
{
  dec->immed_set = true;
  dec->immed = (immed_t)i;
}

static void set_dec_bit (struct decoder_info *dec, unsigned int bit)
{
  dec->bit_set = true;
  dec->bit = (bit_t)bit;
}

static void set_dec_ioaddr (struct decoder_info *dec, unsigned int ioaddr)
{
  dec->ioaddr_set = true;
  dec->ioaddr = (io_data_addr_t) ioaddr;
}

static void set_dec_target (struct decoder_info *dec, unsigned int k)
{
  dec->target_set = true;
  dec->target = (code_addr_t) k;
}

static void set_dec_disp (struct decoder_info *dec, unsigned int q)
{
  dec->disp_set = true;
  dec->disp = (disp_t)q;
}

static bool is_ldi (unsigned short insn, int r, int *immed)
{
  if (
      ((insn & 0xf000) == 0xe000)
      && (((insn & 0x00f0) >> 4) == r-16)
      ) {
    *immed = ((insn & 0x0f0f) >> 4) + (insn & 0x000f);
    return true;
  } else {
    return false;
  }
}

static bool is_out (unsigned short insn, int reg, io_data_addr_t io)
{
  if (
      ((insn & 0xf800) == 0xb800) // out
      && (((insn & 0x01f0) >> 4) == reg) 
      && ((((insn & 0x0600) >> 5) + (unsigned int)(insn & 0x000f)) == (unsigned int)io) 
      ) {
    return true;
  } else {
    return false;
  }
}

static bool check_for_set_stack (code_addr_t addr)
{ 
  int r28val;
  int r29val;
  
  if (!is_ldi (get_insn_short (addr),   28, &r28val)) return false;
  if (!is_ldi (get_insn_short (addr+2), 29, &r29val)) return false;  
  if (!is_out (get_insn_short (addr+4), 29, SPH_ADDR)) return false;
  if (!is_out (get_insn_short (addr+6), 28, SPL_ADDR)) return false;

  all_code[addr]->decode.raw_opcode2_set = false;
  all_code[addr]->decode.opcode = AVR_SET_STACK_PTR;
  all_code[addr]->decode.K_set = true;
  all_code[addr]->decode.K = (r29val << 8) + r28val;
  all_code[addr]->decode.op_type = 47;
  all_code[addr]->decode.size = 8;
  
  for (unsigned int i=0; i<all_code[addr]->decode.size; i++) {
    sp_consistent[addr+i] = false;
  }
  
  init_stack_ptr = (data_addr_t) all_code[addr]->decode.K;

  if (!quiet) {
    if (init_stack_ptr == SRAM_TOP) {
      if (DEBUG_DECODE) {
	printf ("stack ptr initialized to expected value %x\n",
		init_stack_ptr);
      }
    } else {
      printf ("WOAH -- stack ptr initialized to %x (expected %x)\n", 
	      init_stack_ptr,
	      SRAM_TOP);
    }
  }

  return true;
}

static bool is_in (unsigned short insn, int r, io_data_addr_t io)
{
  if (
      ((insn & 0xf800) == 0xb000) // in
      && (((insn & 0x01f0) >> 4) == r) 
      && ((((insn & 0x0600) >> 5) + (unsigned int)(insn & 0x000f)) == (unsigned int)io) 
      ) {
    return true;
  } else {
    return false;
  }
}

static bool is_cli (unsigned short insn)
{
  if ((insn & 0xffff) == 0x94f8) {
    return true;
  } else {
    return false;
  }
}

static bool is_sei (unsigned short insn)
{
  if ((insn & 0xffff) == 0x9478) {
    return true;
  } else {
    return false;
  }
}

static bool is_sbiw (unsigned short insn, int r, int *immed)
{
  if (
      ((insn & 0xff00) == 0x9700) 
      && ((24 + (2 * ((insn & 0x0030) >> 4))) == r)
      ) {
    *immed = ((insn & 0x00c0) >> 2) + (insn & 0x000f);
    return true;
  } else {
    return false;
  }
}

static bool is_adiw (unsigned short insn, int r, int *immed)
{
  if (
      ((insn & 0xff00) == 0x9600) 
      && ((24 + (2 * ((insn & 0x0030) >> 4))) == r)
      ) {
    *immed = ((insn & 0x00c0) >> 2) + (insn & 0x000f);
    return true;
  } else {
    return false;
  }
}

static bool is_subi (unsigned short insn, int r, int *immed)
{
  if (((insn & 0xf000) == 0x5000)
      && (((insn & 0x00f0) >> 4) + 16 == r)) {
    *immed = (((insn & 0x0f00) >> 4) + (insn & 0x000f));
    return true;
  } else {
    return false;
  }
}

static bool is_sbci (unsigned short insn, int r, int *immed)
{
  if (((insn & 0xf000) == 0x4000)
      && (((insn & 0x00f0) >> 4) + 16 == r)) {
    *immed = (((insn & 0x0f00) >> 4) + (insn & 0x000f));
    return true;
  } else {
    return false;
  }
}

static bool is_sub (unsigned short insn, int rd, int rr)
{
  if ((insn & 0xfc00) != 0x1800) return false;
  int myrd = (insn & 0x01f0) >> 4;
  int myrr = (insn & 0x000f) + ((insn & 0x0200) >> 5);
  if (rd != myrd) return false;
  if (rr != myrr) return false;
  return true;
}

static bool is_add (unsigned short insn, int rd, int rr)
{
  if ((insn & 0xfc00) != 0x0c00) return false;
  int myrd = (insn & 0x01f0) >> 4;
  int myrr = (insn & 0x000f) + ((insn & 0x0200) >> 5);
  if (rd != myrd) return false;
  if (rr != myrr) return false;
  return true;
}

static bool is_adc (unsigned short insn, int rd, int rr)
{
  if ((insn & 0xfc00) != 0x1c00) return false;
  int myrd = (insn & 0x01f0) >> 4;
  int myrr = (insn & 0x000f) + ((insn & 0x0200) >> 5);
  if (rd != myrd) return false;
  if (rr != myrr) return false;
  return true;
}

static bool is_sbc (unsigned short insn, int rd, int rr)
{
  if ((insn & 0xfc00) != 0x0800) return false;
  int myrd = (insn & 0x01f0) >> 4;
  int myrr = (insn & 0x000f) + ((insn & 0x0200) >> 5);
  if (rd != myrd) return false;
  if (rr != myrr) return false;
  return true;
}

static void make_inc_stack_insn (code_addr_t addr, int add_amt, unsigned int size)
{
  all_code[addr]->decode.stack_opcode = AVR_INC_STACK_PTR;
  assert (all_code[addr]->decode.stack_delta == SD_UNKNOWN);
  all_code[addr]->decode.stack_delta = add_amt;
  all_code[addr]->decode.stack_opcode_size = size;
  for (unsigned int i=0; i<size; i++) {
    sp_consistent[addr+i] = false;
  }
}

static void make_dec_stack_insn (code_addr_t addr, int sub_amt, unsigned int size)
{
  all_code[addr]->decode.stack_opcode = AVR_DEC_STACK_PTR;
  assert (all_code[addr]->decode.stack_delta == SD_UNKNOWN);
  all_code[addr]->decode.stack_delta = sub_amt;
  all_code[addr]->decode.stack_opcode_size = size;
  for (unsigned int i=0; i<size; i++) {
    sp_consistent[addr+i] = false;
  }
}

static bool check_for_inc1_stack (code_addr_t addr)
{
  int add_amt;

  if (!is_adiw (get_insn_short (addr),    28, &add_amt)) return false;
  if (!is_in   (get_insn_short (addr+2),   0, SREG_ADDR)) return false;
  if (!is_cli  (get_insn_short (addr+4))) return false;
  if (!is_out  (get_insn_short (addr+6),  29, SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+8),   0, SREG_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+10), 28, SPL_ADDR)) return false;

  make_inc_stack_insn (addr, add_amt, 12);
  return true;
}

static bool check_for_inc2_stack (code_addr_t addr)
{
  int add_amt;

  if (!is_adiw (get_insn_short (addr), 24, &add_amt)) return false;
  if (!is_in   (get_insn_short (addr+2), 0, SREG_ADDR)) return false;
  if (!is_cli  (get_insn_short (addr+4))) return false;
  if (!is_out  (get_insn_short (addr+6), 25, SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+8), 0, SREG_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+10), 24, SPL_ADDR)) return false;

  make_inc_stack_insn (addr, add_amt, 12);
  return true;
}

static bool check_for_inc3_stack (code_addr_t addr)
{
  int add_amt;

  if (!is_adiw (get_insn_short (addr), 28, &add_amt)) return false;
  if (!is_cli  (get_insn_short (addr+2))) return false;
  if (!is_out  (get_insn_short (addr+4), 29, SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+6), 28, SPL_ADDR)) return false;

  make_inc_stack_insn (addr, add_amt, 8);
  return true;
}

static bool check_for_inc4_stack (code_addr_t addr)
{
  if (!is_add  (get_insn_short (addr),     28,        30)) return false;
  if (!is_adc  (get_insn_short (addr+2),   29,         1)) return false;
  if (!is_in   (get_insn_short (addr+4),    0, SREG_ADDR)) return false;
  if (!is_cli  (get_insn_short (addr+6)))                  return false;
  if (!is_out  (get_insn_short (addr+8),   29,  SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+10),   0, SREG_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+12),  28,  SPL_ADDR)) return false;

  make_inc_stack_insn (addr, INC_UNK_1, 14);
  return true;
}

static bool check_for_inc5_stack (code_addr_t addr)
{
  int add_amt1, add_amt2;

  // add by subtracting, a brilliant idiom...
  if (!is_subi (get_insn_short (addr),     28, &add_amt1)) return false;
  if (!is_sbci (get_insn_short (addr+2),   29, &add_amt2)) return false;
  if (!is_in   (get_insn_short (addr+4),    0, SREG_ADDR)) return false;
  if (!is_cli  (get_insn_short (addr+6)))                  return false;
  if (!is_out  (get_insn_short (addr+8),   29,  SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+10),   0, SREG_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+12),  28,  SPL_ADDR)) return false;

  int add_amt = 65536 - (256*add_amt2) - add_amt1;
  make_inc_stack_insn (addr, add_amt, 14);
  return true;
}

static bool check_for_dec1_stack (code_addr_t addr)
{ 
  int sub_amt;

  if (!is_in   (get_insn_short (addr),   28, SPL_ADDR))  return false;
  if (!is_in   (get_insn_short (addr+2), 29, SPH_ADDR))  return false;
  if (!is_sbiw (get_insn_short (addr+4), 28, &sub_amt))  return false;
  if (!is_in   (get_insn_short (addr+6), 0, SREG_ADDR))  return false;
  if (!is_cli  (get_insn_short (addr+8)))                return false;
  if (!is_out  (get_insn_short (addr+10), 29, SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+12), 0, SREG_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+14), 28, SPL_ADDR)) return false;
  make_dec_stack_insn (addr, sub_amt, 16);
  
  return true;
}

static bool check_for_dec2_stack (code_addr_t addr)
{ 
  int sub_amt;

  if (!is_in (get_insn_short (addr),   28, SPL_ADDR)) return false;
  if (!is_in (get_insn_short (addr+2), 29, SPH_ADDR)) return false;
  if (!is_sbiw (get_insn_short (addr+4), 28, &sub_amt)) return false;
  if (!is_out (get_insn_short (addr+6), 29, SPH_ADDR)) return false;
  if (!is_out (get_insn_short (addr+8), 28, SPL_ADDR)) return false;
 
  make_dec_stack_insn (addr, sub_amt, 10);
  return true;
}

static bool check_for_dec3_stack (code_addr_t addr)
{ 
  int sub_amt1, sub_amt2;

  if (!is_in   (get_insn_short (addr),   28, SPL_ADDR)) return false;
  if (!is_in   (get_insn_short (addr+2), 29, SPH_ADDR)) return false;
  if (!is_subi (get_insn_short (addr+4), 28, &sub_amt1)) return false;
  if (!is_sbci (get_insn_short (addr+6), 29, &sub_amt2)) return false;
  if (!is_in   (get_insn_short (addr+8), 0, SREG_ADDR)) return false;
  if (!is_cli  (get_insn_short (addr+10))) return false;
  if (!is_out  (get_insn_short (addr+12), 29, SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+14), 0, SREG_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+16), 28, SPL_ADDR)) return false;
  
  make_dec_stack_insn (addr, sub_amt1 + (256 * sub_amt2), 18);
  return true;
}

static bool check_for_dec4_stack (code_addr_t addr)
{ 
  int sub_amt;

  if (!is_in (get_insn_short (addr),   28, SPL_ADDR)) return false;
  if (!is_in (get_insn_short (addr+2), 29, SPH_ADDR)) return false;
  if (!is_sbiw (get_insn_short (addr+4), 28, &sub_amt)) return false;
  if (!is_cli (get_insn_short (addr+6))) return false;
  if (!is_out (get_insn_short (addr+8), 29, SPH_ADDR)) return false;
  if (!is_sei (get_insn_short (addr+10))) return false;
  if (!is_out (get_insn_short (addr+12), 28, SPL_ADDR)) return false;
  
  make_dec_stack_insn (addr, sub_amt, 14);
  return true;
}

static bool check_for_dec5_stack (code_addr_t addr)
{ 
  int sub_amt1, sub_amt2;

  if (!is_in   (get_insn_short (addr),   28, SPL_ADDR)) return false;
  if (!is_in   (get_insn_short (addr+2), 29, SPH_ADDR)) return false;
  if (!is_subi (get_insn_short (addr+4), 28, &sub_amt1)) return false;
  if (!is_sbci (get_insn_short (addr+6), 29, &sub_amt2)) return false;
  if (!is_out  (get_insn_short (addr+8), 29, SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+10), 28, SPL_ADDR)) return false;
  
  make_dec_stack_insn (addr, sub_amt1 + (256 * sub_amt2), 12);
  return true;
}

static bool check_for_dec6_stack (code_addr_t addr)
{ 
  if (!is_in   (get_insn_short (addr),   28, SPL_ADDR)) return false;
  if (!is_in   (get_insn_short (addr+2), 29, SPH_ADDR)) return false;
  if (!is_sub  (get_insn_short (addr+4), 28, 26)) return false;
  if (!is_sbc  (get_insn_short (addr+6), 29, 27)) return false;
  if (!is_in   (get_insn_short (addr+8), 0, SREG_ADDR)) return false;
  if (!is_cli  (get_insn_short (addr+10))) return false;
  if (!is_out  (get_insn_short (addr+12), 29, SPH_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+14), 0, SREG_ADDR)) return false;
  if (!is_out  (get_insn_short (addr+16), 28, SPL_ADDR)) return false;

  make_dec_stack_insn (addr, DEC_UNK_1, 18);
  return true;
}

static void decode_insn (code_addr_t addr, 
			 unsigned short raw_insn, 
			 struct decoder_info *dec) 
{
  /*
   * look for idiomatic "macro instructions" generated by gcc that have
   * a special effect on the stack 
   */

  // but not if we're already in one!
  if (!sp_consistent[addr]) goto done;

  if (check_for_set_stack (addr)) return;
  
  if (check_for_dec1_stack (addr)) goto done;
  if (check_for_dec2_stack (addr)) goto done;
  if (check_for_dec3_stack (addr)) goto done;
  if (check_for_dec4_stack (addr)) goto done;
  if (check_for_dec5_stack (addr)) goto done;
  if (check_for_dec6_stack (addr)) goto done;
  if (check_for_inc1_stack (addr)) goto done;
  if (check_for_inc2_stack (addr)) goto done;
  if (check_for_inc3_stack (addr)) goto done;
  if (check_for_inc4_stack (addr)) goto done;
  if (check_for_inc5_stack (addr)) goto done;
 done:
  ;
  
#include "avr_decode_switch.ci"
}

static void decode_args (struct decoder_info *dec)
{
  switch (dec->op_type) {
    
  case 0:
    break;

  case 1:
    assert (dec->d_set);
    assert (dec->r_set);
    set_dec_dst_reg (dec, dec->d);
    set_dec_src_reg (dec, dec->r);
    break;
    
  case 2:
    assert (dec->K_set);
    assert (dec->d_set);
    {
      unsigned int rd = 2 * dec->d + 24;
      unsigned int K = dec->K;
      set_dec_dst_reg (dec, rd);
      set_dec_immed (dec, K);
    }
    break;

  case 3:
    assert (dec->K_set);
    assert (dec->d_set);
    {
      unsigned int K = dec->K;
      unsigned int rd = dec->d + 16;
      set_dec_dst_reg (dec, rd);
      set_dec_immed (dec, K);
    }
    break;

  case 4:
    assert (dec->d_set);
    set_dec_dst_reg (dec, dec->d);
    break;

  case 5:
    die ();

  case 6:
    assert (dec->d_set);
    assert (dec->b_set);
    set_dec_dst_reg (dec, dec->d);
    set_dec_bit (dec, dec->b);
    break;

  case 7:
    die ();

  case 8:
    assert (dec->k_set);
    {
      unsigned int k = unsigned_to_twos_comp (7, dec->k);
      set_dec_immed (dec, k*2);
      set_dec_target (dec, 2 * (k + (dec->addr/2) + 1));
    }
    break;

  case 9:
    assert (dec->k_set);
    set_dec_target (dec, 2*(dec->k));
    break;

  case 10:
    assert (dec->b_set);
    assert (dec->A_set);
    set_dec_bit (dec, dec->b);
    set_dec_ioaddr (dec, dec->A);
    break;

  case 14:
    assert (dec->d_set);
    assert (dec->k_set);

    {
      unsigned int rd = dec->d;
      unsigned int k = dec->k;

      set_dec_dst_reg (dec, rd);
      set_dec_immed (dec, k);
    }
    break;

  case 15:
    assert (dec->d_set);
    assert (dec->r_set);
    set_dec_dst_reg (dec, dec->d * 2);
    set_dec_src_reg (dec, dec->r * 2);
    break;
    
  case 16:
    assert (dec->r_set);
    assert (dec->d_set);
    set_dec_src_reg (dec, dec->r + 16);
    set_dec_dst_reg (dec, dec->d + 16);
    break;

  case 17:
    assert (dec->A_set);
    assert (dec->r_set);
    set_dec_src_reg (dec, dec->r);
    set_dec_ioaddr (dec, dec->A);
    break;

  case 18:
    assert (dec->k_set);
    {
      unsigned int k = unsigned_to_twos_comp (12, dec->k);
      set_dec_immed (dec, 2*k);
      set_dec_target (dec, 2*(k + dec->addr/2 + 1));
    }
    break;

  case 20:
    assert (dec->r_set);
    set_dec_src_reg (dec, dec->r);
    break;

  case 21:
    assert (dec->r_set);
    assert (dec->b_set);
    set_dec_src_reg (dec, dec->r);
    set_dec_bit (dec, dec->b);
    break;

  case 22:
    assert (dec->A_set);
    assert (dec->d_set);
    set_dec_dst_reg (dec, dec->d);
    set_dec_ioaddr (dec, dec->A);
    break;

  case 23:
    assert (dec->r_set);
    assert (dec->k_set);
    set_dec_src_reg (dec, dec->r);
    set_dec_immed (dec, dec->k);
    break;

  case 25:
  case 26:
  case 27:
  case 28:
  case 29:
  case 30:
  case 31:
  case 32:
  case 33:
  case 34:
  case 35:
    assert (dec->d_set);
    set_dec_dst_reg (dec, dec->d);
    if (dec->op_type == 31 ||
	dec->op_type == 35) {
      assert (dec->q_set);
      set_dec_disp (dec, dec->q);
    }
    break;

  case 36:
  case 37:
  case 38:
  case 39:
  case 40:
  case 41:
  case 42:
  case 43:
  case 44:
  case 45:
  case 46:
    assert (dec->r_set);
    set_dec_src_reg (dec, dec->r);
    if (dec->op_type == 42 ||
	dec->op_type == 46) {
      assert (dec->q_set);
      set_dec_disp (dec, dec->q);
    }
    break;

  case 47:
  case 48:
    break;

  default:
    printf ("\n\noops -- unhandled opcode %s\n", opcode_to_string (dec->opcode));
    die ();
  }
}

static void print_args (const struct decoder_info *dec, FILE *outf)
{
  switch (dec->op_type) {

  case 0:
    break;

  case 1:
  case 15:
  case 16:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    fprintf (outf, "\tr%d, r%d", 
	    dec->dst_reg, 
	    dec->src_reg);
    break;
    
  case 2:
    assert (dec->dst_reg_set);
    assert (dec->immed_set);    
    fprintf (outf, "\tr%d, 0x%02x\t; %u", 
	    dec->dst_reg,
	    dec->immed, dec->immed);
    break;

  case 3:
    assert (dec->dst_reg_set);
    assert (dec->immed_set);    
    fprintf (outf, "\tr%d, 0x%02X\t; %u", 
	    dec->dst_reg,
	    dec->immed, dec->immed);
    break;

  case 4:
    assert (dec->dst_reg_set);
    fprintf (outf, "\tr%d", 
	    dec->dst_reg);
    break;

  case 5:
    die ();

  case 6:
    assert (dec->dst_reg_set);
    assert (dec->bit_set);
    fprintf (outf, "\tr%d, %d", dec->dst_reg, dec->bit);
    break;

  case 7:
    die ();

  case 8:
    assert (dec->target_set);
    assert (dec->immed_set);
    fprintf (outf, "\t.%+02d\t\t; 0x%x",
	    dec->immed, 
	    dec->target);
    break;
    
  case 9:
    assert (dec->target_set);
    // N.B. deviates from objdump output
    decode_addr (dec->target);
    fprintf (outf, "\t0x%x\t\t; %s", 
	    dec->target,
	     (all_code[dec->target]->decode.label == NULL) ? "" :
	     all_code[dec->target]->decode.label);
    break;

  case 10:
    assert (dec->bit_set);
    assert (dec->ioaddr_set);
    fprintf (outf, "\t0x%02x, %d\t; %d", dec->ioaddr, dec->bit, dec->ioaddr);
    break;

  case 14:
    assert (dec->dst_reg_set);
    assert (dec->immed_set);
    fprintf (outf, "\tr%d, 0x%04X", dec->dst_reg, dec->immed);
    break;

  case 17:
    assert (dec->ioaddr_set);
    assert (dec->src_reg_set);
    fprintf (outf, "\t0x%02x, r%d\t; %d", dec->ioaddr, dec->src_reg, dec->ioaddr);
    break;

  case 18:
    assert (dec->target_set);
    assert (dec->immed_set);
    fprintf (outf, "\t.%+d\t\t; 0x%x", dec->immed, dec->target);
    break;

  case 20:
    assert (dec->src_reg_set);
    fprintf (outf, "\tr%d", 
	    dec->src_reg);
    break;

  case 21:
    assert (dec->src_reg_set);
    assert (dec->bit_set);
    fprintf (outf, "\tr%d, %d", dec->src_reg, dec->bit);
    break;

  case 22:
    assert (dec->ioaddr_set);
    assert (dec->dst_reg_set);
    fprintf (outf, "\tr%d, 0x%02x\t; %d", dec->dst_reg, dec->ioaddr, dec->ioaddr);
    break;

  case 23:
    assert (dec->src_reg_set);
    assert (dec->immed_set);
    fprintf (outf, "\t0x%04X, r%d", dec->immed, dec->src_reg);
    break;

  case 25:
  case 26:
  case 27:
  case 28:
  case 29:
  case 30:
  case 31:
  case 32:
  case 33:
  case 34:
  case 35:
    assert (dec->dst_reg_set);
    {
      reg_t rd = dec->dst_reg;

      switch (dec->op_type) {
      case 25:
	fprintf (outf, "\tr%d, X", rd);
	break;
      case 26:
	fprintf (outf, "\tr%d, X+", rd);
	break;
      case 27:
	fprintf (outf, "\tr%d, -X", rd);
	break;
      case 28:
	fprintf (outf, "\tr%d, Y", rd);
	break;
      case 29:
	fprintf (outf, "\tr%d, Y+", rd);
	break;
      case 30:
	fprintf (outf, "\tr%d, -Y", rd);
	break;
      case 31:
	assert (dec->disp_set);
	fprintf (outf, "\tr%d, Y+%d\t; 0x%02x", rd, dec->disp, dec->disp);
	break;
      case 32:
	fprintf (outf, "\tr%d, Z", rd);
	break;
      case 33:
	fprintf (outf, "\tr%d, Z+", rd);
	break;
      case 34:
	fprintf (outf, "\tr%d, -Z", rd);
	break;
      case 35:
	assert (dec->disp_set);
	fprintf (outf, "\tr%d, Z+%d\t; 0x%02x", rd, dec->disp, dec->disp);
	break;
      default:
	die ();
      }
    }
    break;

  case 36:
  case 37:
  case 38:
  case 39:
  case 40:
  case 41:
  case 42:
  case 43:
  case 44:
  case 45:
  case 46:
    assert (dec->src_reg_set);
    {
      reg_t rr = dec->src_reg;

      switch (dec->op_type) {
      case 36:
	fprintf (outf, "\tX, r%d", rr);
	break;
      case 37:
	fprintf (outf, "\tX+, r%d", rr);
	break;
      case 38:
	fprintf (outf, "\t-X, r%d", rr);
	break;
      case 39:
	fprintf (outf, "\tY, r%d", rr);
	break;
      case 40:
	fprintf (outf, "\tY+, r%d", rr);
	break;
      case 41:
	fprintf (outf, "\t-Y, r%d", rr);
	break;
      case 42:
	assert (dec->disp_set);
	fprintf (outf, "\tY+%d, r%d\t; 0x%02x", dec->disp, rr, dec->disp);
	break;
      case 43:
	fprintf (outf, "\tZ, r%d", rr);
	break;
      case 44:
	fprintf (outf, "\tZ+, r%d", rr);
	break;
      case 45:
	fprintf (outf, "\t-Z, r%d", rr);
	break;
      case 46:
	assert (dec->disp_set);
	fprintf (outf, "\tZ+%d, r%d\t; 0x%02x", dec->disp, rr, dec->disp);
	break;
      default:
	die ();
      }
    }
    break;

  case 47:
  case 48:
    die ();

  default:
    fprintf (outf, "\n\noops -- unhandled opcode %s\n", opcode_to_string (dec->opcode));
    die ();
  }
}

static void set_stack_delta (struct decoder_info *dec)
{
  int sd;

  if (dec->stack_opcode != NO_STACK_OPCODE) {
    return;
  }

  switch (dec->opcode) {

  case AVR_call:
  case AVR_rcall:
  case AVR_icall:
    sd = PC_SIZE;
    break;

  case AVR_ret:
  case AVR_reti:    
    sd = -PC_SIZE;
    break;

  case AVR_eicall:
    sd = 3;  // correct on all AVR chips, according to the manual
    break;

  case AVR_push:
    sd = 1;
    break;

  case AVR_pop:
    sd = -1;
    break;

  default:
    sd = 0;
    break;
  }

  /*
   * account for implicit pushes at interrupt vectors (except vector 0
   * which pushes nothing since SP is not set when it it jumped to)
   */
  for (unsigned int i=0; i<platform->num_vectors; i++) {
    if (dec->addr == platform->int_vecs[i].start_addr &&
	platform->int_vecs[i].start_addr != (code_addr_t)0) {
      sd += PC_SIZE;
    }
  }

  assert (dec->stack_delta == SD_UNKNOWN);
  dec->stack_delta = sd;
}

// Return costs for all instructions that don't branch
cost_t avr_cost (enum AVR_OPCODE opcode) {
  switch (opcode) {
    // All of these instructions should have simple control flow
    // (i.e., transfer control to immediate successor)
  case AVR_adc : return 1;
  case AVR_add : return 1;
  case AVR_adiw : return 2;
  case AVR_and : return 1;
  case AVR_andi : return 1;
  case AVR_asr : return 1;
  case AVR_bld  : return 1;
  case AVR_break : return 1;
  case AVR_bst : return 1;
  case AVR_cbi : return 2;
  case AVR_clh : return 1;
  case AVR_cli : return 1;
  case AVR_cln : return 1;
  case AVR_cls : return 1;
  case AVR_clt : return 1;
  case AVR_clv : return 1;
  case AVR_clz : return 1;
  case AVR_com : return 1;
  case AVR_cp : return 1;
  case AVR_cpc : return 1;
  case AVR_cpi : return 1;
  case AVR_cpse : return 1;
  case AVR_dec : return 1;
  case AVR_elpm1 : return 3;
  case AVR_elpm2 : return 3;
  case AVR_elpm3 : return 3;
  case AVR_eor : return 1;
  case AVR_fmul : return 2;
  case AVR_fmuls : return 2;
  case AVR_fmulsu : return 2;
  case AVR_in : return 1;
  case AVR_inc : return 1;
  case AVR_ldi : return 1;
  case AVR_lds : return 2;
  case AVR_lpm1 : return 3;
  case AVR_lpm2 : return 3;
  case AVR_lpm3 : return 3;
  case AVR_lsr : return 1;
  case AVR_mov : return 1;
  case AVR_movw : return 1;
  case AVR_mul : return 2;
  case AVR_muls : return 2;
  case AVR_mulsu : return 2;
  case AVR_neg : return 1;
  case AVR_nop : return 1;
  case AVR_or : return 1;
  case AVR_ori : return 1;
  case AVR_out : return 1;
  case AVR_pop : return 2;
  case AVR_push : return 2;
  case AVR_ror : return 1;
  case AVR_sbc : return 1;
  case AVR_sbci : return 1;
  case AVR_sbi : return 2;
  case AVR_sbiw : return 2;
  case AVR_sec : return 1;
  case AVR_seh : return 1;
  case AVR_sei : return 1;
  case AVR_sen : return 1;
  case AVR_ses : return 1;
  case AVR_set : return 1;
  case AVR_sev : return 1;
  case AVR_sez : return 1;
  case AVR_sleep : return 1;
  case AVR_spm : return 1;
  case AVR_sts : return 2;
  case AVR_sub : return 1;
  case AVR_subi : return 1;
  case AVR_swap : return 1;
  case AVR_wdr : return 1;
  case AVR_ld1x : return 2;
  case AVR_ld2x : return 2;
  case AVR_ld3x : return 2;
  case AVR_ld1y : return 2;
  case AVR_ld2y : return 2;
  case AVR_ld3y : return 2;
  case AVR_ldd1y : return 2;
  case AVR_ld1z : return 2;
  case AVR_ld2z : return 2;
  case AVR_ld3z : return 2;
  case AVR_ldd1z : return 2;
  case AVR_st1x : return 2;
  case AVR_st2x : return 2;
  case AVR_st3x : return 2;
  case AVR_st1y : return 2;
  case AVR_st2y : return 2;
  case AVR_st3y : return 2;
  case AVR_std1y : return 2;
  case AVR_st1z : return 2;
  case AVR_st2z : return 2;
  case AVR_st3z : return 2;
  case AVR_std1z : return 2;
  case AVR_SET_STACK_PTR : return 10;

    // All these instructions have complex control flow (so they're not
    // handled here).
  case AVR_breq : 
  case AVR_brge : 
  case AVR_brne : 
  case AVR_brcs : 
  case AVR_brcc : 
  case AVR_brlt : 
  case AVR_brhc : 
  case AVR_brhs : 
  case AVR_brid : 
  case AVR_brie : 
  case AVR_brmi : 
  case AVR_brpl : 
  case AVR_brtc : 
  case AVR_brts : 
  case AVR_brvc : 
  case AVR_brvs : 
  case AVR_brbc : 
  case AVR_brbs : 
  case AVR_call : 
  case AVR_eicall:
  case AVR_eijmp: 
  case AVR_icall: 
  case AVR_ijmp : 
  case AVR_jmp  : 
  case AVR_rcall: 
  case AVR_ret  : 
  case AVR_reti : 
  case AVR_rjmp : 
  case AVR_sbic : 
  case AVR_sbis : 
  case AVR_sbrc : 
  case AVR_sbrs : 
    die();

  case AVR_UNKNOWN : die();
  }
  die();
};

static avr_insn *allocate_insn (void)
{
  avr_insn *insn;
  insn = (avr_insn *) xmalloc (sizeof (avr_insn));
  memset (insn, 0, sizeof (avr_insn));
  insn->control.index_taken = NO_IDX;
  insn->control.index_not_taken = NO_IDX;
#if WCET_ANALYSIS
  insn->loop_end = NO_CODE_ADDR;
  insn->exn_count = NEG_INFINITY;
#endif
  return insn;
}

void undecode_addr (code_addr_t addr)
{
  if (all_code[addr] != NULL) {
    free (all_code[addr]);
    all_code[addr] = NULL;
  }
}

void decode_addr (code_addr_t addr)
{
  unsigned short raw_insn;
  struct decoder_info *dec;

  if (DEBUG_DECODE) printf ("decoding %x\n", addr);

  // make it harmless to call this routine multiple times
  if (all_code[addr] != NULL) {
    return;
  }

  all_code[addr] = allocate_insn ();
  dec = &all_code[addr]->decode;

  raw_insn = get_insn_short (addr);
  
  dec->opcode = AVR_UNKNOWN;
  dec->addr = addr;
  dec->raw_opcode = raw_insn;

  dec->stack_delta = SD_UNKNOWN;
  dec->stack_opcode = NO_STACK_OPCODE;
  dec->stack_opcode_size = -1;

  decode_insn (addr, raw_insn, dec);

  if (DEBUG_DECODE) {
    printf ("addr = %x,  insn = %s %04x,  size = %d\n", 
	    addr, 
	    opcode_to_string (dec->opcode), 
	    raw_insn,
	    dec->size);
  }

  assert (dec->size > 0);

  if (dec->opcode == AVR_UNKNOWN) {
    free (all_code[addr]);
    all_code[addr] = NULL;
    return;
  }

  decode_args (dec);
  set_stack_delta (dec);
  assert (dec->stack_delta != SD_UNKNOWN);
  dec->label = lookup_text_symbol (dec->addr);
}

static unsigned char low_opcode_byte (unsigned short op)
{
  return op & 0xff;
}

static unsigned char high_opcode_byte (unsigned short op)
{
  return op >> 8;
}

    
void avr_disasm_insn (code_addr_t addr, FILE* outf)
{
  const struct decoder_info *dec = &all_code[addr]->decode;

  fprintf (outf, "%s  ", opcode_to_string (dec->opcode));
    
  switch (dec->opcode) {
      
  case AVR_SET_STACK_PTR:
    fprintf (outf, "\t0x%04x\t; %u", dec->K, dec->K);
    break;
    
  default:
    print_args (dec, outf);
  }
  
}

bool is_live_instr(code_addr_t i) {
  if (all_code[i] == NULL) return false;
  for (code_ctxt_t j=(code_ctxt_t)0; j < (code_ctxt_t)all_code[i]->state_cnt; j++) {
    if (all_code[i]->states[j].visit_cnt > 0) return true;
  }
  return false;
}

static void disasm_insn (code_addr_t addr, FILE *outf, ostream &os)
{
  decode_addr (addr);

  const struct decoder_info *dec = &all_code[addr]->decode;
  const struct control_info *ctl = &all_code[addr]->control;

  if (all_code[addr] == NULL) {
    fprintf (outf, ".word\t0x%04x\t\t; NOT A KNOWN INSTRUCTION\n", 
	     get_insn_short (addr));
    return;
  }

  if (dec->label != NULL) {
    fprintf (outf, "\n");
    fprintf (outf, "%08x <%s>:", addr, dec->label);
    fprintf (outf, "\n");
  }

  // N.B. these make stacktool output differ from objdump

  if (true) {
    fprintf (outf, "%s ", 
	     all_code[addr]->wc_path ? "X" : (all_code[addr]->touched ? "*" : " "));
  }

  if (true) {
    fprintf (outf, "%d %d ", 
	    ctl->edges_in.len, 
	    ctl->edges_out.len);
  } 
  if (false) {
    fprintf (outf, "%2d ", 
	    dec->stack_delta);
  }  

  if (true) {
    if (all_code[addr]->touched) {
      fprintf (outf, "%2d ", 
	      all_code[addr]->wc_stack_depth);
    } else {
      fprintf (outf, "   ");
    }
  }

  if (true) {
    print_brief_dataflow (addr, outf);
  }
  
  fprintf (outf, "%8x:\t", dec->addr);  
  fprintf (outf, "%02x %02x ", 
	  low_opcode_byte (dec->raw_opcode), 
	  high_opcode_byte (dec->raw_opcode));
  
  if (dec->raw_opcode2_set) {
    fprintf (outf, "%02x %02x ",
	    low_opcode_byte (dec->raw_opcode2),
	    high_opcode_byte (dec->raw_opcode2));
  } else {
    fprintf (outf, "     ");
  }
  
  fprintf (outf, "\t");
  avr_disasm_insn (addr,outf);
  fprintf (outf, "\n");

  if (verbose) {
    print_verbose_dataflow (addr, outf, os);
  }
}

void disassemble (const char *fn)
{
  code_addr_t i = 0;
  FILE *disasm_file;

  // fprintf(stderr,"Writing disassembly to '%s'\n",fn);
  disasm_file = fopen (fn, "w");
  assert (disasm_file != NULL);

  // The following grunge converts a C FILE* to a C++ ostream.  Blech!
  __gnu_cxx::stdio_filebuf<char,std::char_traits<char> > buf(disasm_file,ios::out|ios::trunc,0);
  ostream os(&buf);

  while ((unsigned int)i<code_size) {
    if (all_code[i] == NULL) {
      i+=2;
      continue;
    }
    if (is_live_instr(i)) {
      disasm_insn (i, disasm_file, os);
    }
    i = nextPC(i);
  }

  (void) fclose (disasm_file);
}

void find_counts (void)
{
  unsigned int known = 0, unknown = 0;
  unsigned int total_cnt;
  code_addr_t i;
  long long int bits = 0, sreg_bits = 0;
  long long int num_states = 0;

  i=0;
  while ((unsigned int)i<code_size) {

    if (all_code[i] == NULL) {
      i+=2;
      continue;
    }

    bool good = true;
    bool non_trivial = false; // are there any non-triv states at this address?

    decode_addr (i);
  
    for (code_ctxt_t j=(code_ctxt_t)0; j < (code_ctxt_t)all_code[i]->state_cnt; j++) {
      
      state_inst *tmp_state = &all_code[i]->states[j];
      
      if (tmp_state->visit_cnt > 0) {

        non_trivial = true;
        sreg_bits += count_bits(tmp_state->mem.core[io_to_mem(SREG_ADDR)].known);

        bits += count_bits(&tmp_state->mem);
        num_states++;

        if ((tmp_state->mem.core[io_to_mem(SREG_ADDR)].known & 0x80) == 0x0) {
          good = false;
        }
      }
    }
    if (non_trivial) { // don't count insns we never visit
      if (good) {
        known++;
      } else {
        unknown++;
      }
    }
    else {
      //      cout << "Address " << hex << i << " is trivial" << endl;
    }
    
    i = nextPC(i);
  }

  total_cnt = known + unknown;
  if (total_cnt > 0) {
    double average = bits / (double) num_states;
    int bits_per_state = 8 * (platform->regs.len + 1);
    printf ("%d insns, %lld states, %lld (%.2f%%) known bits, %.2f%% SREG known, %d (%.2f%%) unknown I masks\n",
	    total_cnt, num_states,
            bits,    100.0 * average / (double)bits_per_state,
            100 * sreg_bits / (double) num_states / 8,
            unknown, 100.0 * unknown / (double)total_cnt
            );
#if VERBOSE_CTRL_FLOW
    printf ("there are %d bits in a machine state\n", bits_per_state);
#endif
    
  }
}
