#include "stacktool.h"

static const mem_loc_t MEM_UNKNOWN = mem_loc_t::bottom;
static const mem_loc_t MEM_ZERO    = mem_loc_t(0);

// when an interrupt fires, all we know is that interrupts are
// disabled
static const mem_loc_t INIT_SREG   = mem_loc_t(0x00,0x80);

///////////////////////////// ATmega 16 ///////////////////////////////

const struct avr_platform atmega16 = {
  "atmega16",
  21,              // vectors
  {
    { "RESET",        0x00, NO_BIT,      NO_BIT      },
    { "INT0",         0x04, { 0x5b, 6 }, { 0x5a, 6 } },
    { "INT1",         0x08, { 0x5b, 7 }, { 0x5a, 7 } },
    { "TIMER2_COMP",  0x0c, { 0x59, 7 }, { 0x58, 7 } },
    { "TIMER2_OVF",   0x10, { 0x59, 6 }, { 0x58, 6 } },
    { "TIMER1_CAPT",  0x14, { 0x59, 5 }, { 0x58, 5 } },
    { "TIMER1_COMPA", 0x18, { 0x59, 4 }, { 0x58, 4 } },
    { "TIMER1_COMPB", 0x1c, { 0x59, 3 }, { 0x58, 3 } },
    { "TIMER1_OVF",   0x20, { 0x59, 2 }, { 0x58, 2 } },
    { "TIMER0_OVF",   0x24, { 0x59, 0 }, { 0x58, 0 } },
    { "SPI_STC",      0x28, { 0x2d, 7 }, NO_BIT      },
    { "UART_RX",      0x2c, { 0x2a, 7 }, { 0x2b, 7 } },
    { "UART_UDRE",    0x30, { 0x2a, 5 }, { 0x2b, 5 } },
    { "UART_TX",      0x34, { 0x2a, 6 }, { 0x2b, 6 } },
    { "ADC",          0x38, { 0x26, 3 }, { 0x26, 4 } },
    { "EE_READY",     0x3c, { 0x3c, 3 }, NO_BIT      },
    { "ANA_COMP",     0x40, { 0x28, 3 }, { 0x28, 4 } },
    { "TWI",          0x44, { 0x57, 0 }, { 0x56, 7 } },
    { "INT2",         0x48, { 0x5b, 5 }, { 0x5a, 5 } },
    { "TIMER0_COMP",  0x4c, { 0x59, 1 }, { 0x58, 1 } },
    { "SPM_READY",    0x50, { 0x57, 7 }, NO_BIT      },
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT
  },
  96,              // state size
  { 0x00,  0x20 }, //   32 bytes registers
  { 0x20,  0x40 }, //   64 bytes io
  { 0x60,  0x00 }, //    0 bytes extended io
  { 0x60, 0x400 }, // 1024 bytes SRAM
};

#if 1
///////////////////////////// ATmega 163 //////////////////////////////

// FIXME -- check

const struct avr_platform atmega163 = {
  "atmega163",
  18,              // vectors
  { 
    { "RESET",        0x00, NO_BIT,      NO_BIT      },
    { "INT0",         0x04, { 0x5b, 6 }, { 0x5a, 6 } },
    { "INT1",         0x08, { 0x5b, 7 }, { 0x5a, 7 } },
    { "TIMER2_COMP",  0x0c, { 0x59, 7 }, { 0x58, 7 } },
    { "TIMER2_OVF",   0x10, { 0x59, 6 }, { 0x58, 6 } },
    { "TIMER1_CAPT",  0x14, { 0x59, 5 }, { 0x58, 5 } },
    { "TIMER1_COMPA", 0x18, { 0x59, 4 }, { 0x58, 4 } },
    { "TIMER1_COMPB", 0x1c, { 0x59, 3 }, { 0x58, 3 } },
    { "TIMER1_OVF",   0x20, { 0x59, 2 }, { 0x58, 2 } },
    { "TIMER0_OVF",   0x24, { 0x59, 0 }, { 0x58, 0 } },
    { "SPI_STC",      0x28, { 0x2d, 7 }, NO_BIT      },
    { "UART_RX",      0x2c, { 0x2a, 7 }, { 0x2b, 7 } },
    { "UART_UDRE",    0x30, { 0x2a, 5 }, { 0x2b, 5 } },
    { "UART_TX",      0x34, { 0x2a, 6 }, { 0x2b, 6 } },
    { "ADC",          0x38, { 0x26, 3 }, { 0x26, 4 } },
    { "EE_READY",     0x3c, { 0x3c, 3 }, NO_BIT      },
    { "ANA_COMP",     0x40, { 0x28, 3 }, { 0x28, 4 } },
    { "TWI",          0x44, { 0x57, 0 }, { 0x56, 7 } },
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
  },
  96,              // state size
  { 0x00,  0x20 }, //   32 bytes registers
  { 0x20,  0x40 }, //   64 bytes io
  { 0x60,  0x00 }, //    0 bytes extended io
  { 0x60, 0x400 }, // 1024 bytes SRAM
};
#endif

///////////////////////////// ATmega 103 //////////////////////////////

const struct avr_platform atmega103 = {
  "atmega103",
  24,              // vectors
  {
    { "RESET",        0x00, NO_BIT,      NO_BIT      },
    { "INT0",         0x04, { 0x59, 0 }, NO_BIT      },
    { "INT1",         0x08, { 0x59, 1 }, NO_BIT      },
    { "INT2",         0x0c, { 0x59, 2 }, NO_BIT      },
    { "INT3",         0x10, { 0x59, 3 }, NO_BIT      },
    { "INT4",         0x14, { 0x59, 4 }, { 0x58, 4 } },
    { "INT5",         0x18, { 0x59, 5 }, { 0x58, 5 } },
    { "INT6",         0x1c, { 0x59, 6 }, { 0x58, 6 } },
    { "INT7",         0x20, { 0x59, 7 }, { 0x58, 7 } },
    { "TIMER2_COMP",  0x24, { 0x57, 7 }, { 0x56, 7 } },
    { "TIMER2_OVF",   0x28, { 0x57, 6 }, { 0x56, 6 } },
    { "TIMER1_CAPT",  0x2c, { 0x57, 5 }, { 0x56, 5 } },
    { "TIMER1_COMPA", 0x30, { 0x57, 4 }, { 0x56, 4 } },
    { "TIMER1_COMPB", 0x34, { 0x57, 3 }, { 0x56, 3 } },
    { "TIMER1_OVF",   0x38, { 0x57, 2 }, { 0x56, 2 } },
    { "TIMER0_COMP",  0x3c, { 0x57, 1 }, { 0x56, 1 } },
    { "TIMER0_OVF",   0x40, { 0x57, 0 }, { 0x56, 0 } },
    { "SPI_STC",      0x44, { 0x2d, 7 }, NO_BIT      },
    { "UART_RX",      0x48, { 0x2a, 7 }, { 0x2b, 7 } },
    { "UART_UDRE",    0x4c, { 0x2a, 5 }, { 0x2b, 5 } },
    { "UART_TX",      0x50, { 0x2a, 6 }, { 0x2b, 6 } },
    { "ADC",          0x54, { 0x26, 3 }, { 0x26, 4 } },
    { "ANA_COMP",     0x58, { 0x28, 3 }, { 0x28, 4 } },
    { "EE_READY",     0x5c, { 0x3c, 3 }, NO_BIT      },
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
    NO_INT,
  },
  96,              // state size
  { 0x00,  0x20 }, //   32 bytes registers
  { 0x20,  0x40 }, //   64 bytes io
  { 0x60,  0x00 }, //    0 bytes extended io
  { 0x60, 0xfa0 }, // 4000 bytes SRAM
};

///////////////////////////// ATmega 128 //////////////////////////////

const struct avr_platform atmega128 = {
  "atmega128",
  35,                // vectors
  {
    { "RESET",        0x00, NO_BIT,      NO_BIT      },
    { "INT0",         0x04, { 0x59, 0 }, { 0x58, 0 } },
    { "INT1",         0x08, { 0x59, 1 }, { 0x58, 1 } },
    { "INT2",         0x0c, { 0x59, 2 }, { 0x58, 2 } },
    { "INT3",         0x10, { 0x59, 3 }, { 0x58, 3 } },
    { "INT4",         0x14, { 0x59, 4 }, { 0x58, 4 } },
    { "INT5",         0x18, { 0x59, 5 }, { 0x58, 5 } },
    { "INT6",         0x1c, { 0x59, 6 }, { 0x58, 6 } },
    { "INT7",         0x20, { 0x59, 7 }, { 0x58, 7 } },
    { "TIMER2_COMP",  0x24, { 0x57, 7 }, { 0x56, 7 } },
    { "TIMER2_OVF",   0x28, { 0x57, 6 }, { 0x56, 6 } },
    { "TIMER1_CAPT",  0x2c, { 0x57, 5 }, { 0x56, 5 } },
    { "TIMER1_COMPA", 0x30, { 0x57, 4 }, { 0x56, 4 } },
    { "TIMER1_COMPB", 0x34, { 0x57, 3 }, { 0x56, 3 } },
    { "TIMER1_OVF",   0x38, { 0x57, 2 }, { 0x56, 2 } },
    { "TIMER0_COMP",  0x3c, { 0x57, 1 }, { 0x56, 1 } },
    { "TIMER0_OVF",   0x40, { 0x57, 0 }, { 0x56, 0 } },
    { "SPI_STC",      0x44, { 0x2d, 7 }, NO_BIT      },
    { "USART0_RX",    0x48, { 0x2a, 7 }, { 0x2b, 7 } },
    { "USART0_UDRE",  0x4c, { 0x2a, 5 }, { 0x2b, 5 } },
    { "USART0_TX",    0x50, { 0x2a, 6 }, { 0x2b, 6 } },
    { "ADC",          0x54, { 0x26, 3 }, { 0x26, 4 } },
    { "EE_READY",     0x58, { 0x3c, 3 }, NO_BIT      },
    { "ANA_COMP",     0x5c, { 0x28, 3 }, { 0x28, 4 } },
    { "TIMER1_COMPC", 0x60, { 0x7d, 0 }, { 0x7c, 0 } },
    { "TIMER3_CAPT",  0x64, { 0x7d, 5 }, { 0x7c, 5 } },
    { "TIMER3_COMPA", 0x68, { 0x7d, 4 }, { 0x7c, 4 } },
    { "TIMER3_COMPB", 0x6c, { 0x7d, 3 }, { 0x7c, 3 } },
    { "TIMER3_COMPC", 0x70, { 0x7d, 1 }, { 0x7c, 1 } },
    { "TIMER3_OVF",   0x74, { 0x7d, 2 }, { 0x7c, 2 } },
    { "USART1_RX",    0x78, { 0x9a, 7 }, { 0x9a, 7 } },
    { "USART1_UDRE",  0x7c, { 0x9a, 5 }, { 0x9a, 5 } },
    { "USART1_TX",    0x80, { 0x9a, 6 }, { 0x9a, 6 } },
    { "TWI",          0x84, { 0x74, 0 }, { 0x74, 7 } },
    { "SPM_READY",    0x88, { 0x68, 7 }, NO_BIT      },
    NO_INT
  },
  96,                // state size
  {  0x00,   0x20 }, //   32 bytes registers
  {  0x20,   0x40 }, //   64 bytes io
  {  0x60,   0xa0 }, //  160 bytes extended io
  { 0x100, 0x1000 }, // 4096 bytes SRAM
};

///////////////////////////////////////////////////////////////////////

const int NUM_PLATFORMS = 4;

const struct avr_platform *all_platforms[NUM_PLATFORMS] = {
  &atmega16,
  &atmega163,
  &atmega103,
  &atmega128,
};

static const mem_loc_t mem_init = mem_loc_t::bottom;
// static const mem_loc_t mem_init = mem_loc_t::zero;

void initialize_values (const struct avr_platform *p,
                        mem_loc_t m[MAX_STATE_SIZE]) 
{
  for(reg_t i=0; i < p->regs.len; ++i) {
    m[reg_to_mem(i)] = mem_init;
  }

  for(io_data_addr_t i=0; i < p->io.len; ++i) {
    m[io_to_mem(i)] = mem_init;
  }

  for(io_data_addr_t i=0; i < p->extended_io.len; ++i) {
    m[extended_io_to_mem(i)] = mem_init;
  }

  m[io_to_mem(SPH_ADDR)]  = mem_loc_t (INIT_SPH);
  m[io_to_mem(SPL_ADDR)]  = mem_loc_t (INIT_SPL);

  m[io_to_mem(SREG_ADDR)] = INIT_SREG;
};

