#!/usr/bin/perl -w

use strict;

sub create_masks ($) {
    my $bits = shift;
    die if (length ($bits) != 16);

    my $insn_bits = $bits;
    ($insn_bits =~ s/[a-zA-Z]/0/g);
    #print "$insn_bits  ";
    my $insn_hex = unpack ("H4", pack ("B16", $insn_bits));
    #print "$insn_hex  ";

    my $insn_mask = $bits;
    ($insn_mask =~ s/0/\@/g);
    ($insn_mask =~ s/[a-zA-Z]/0/g);
    ($insn_mask =~ s/\@/1/g);
    my $mask_hex = unpack ("H4", pack ("B16", $insn_mask));
    #print "$mask_hex  ";
    #print "\n";

    return ("0x$insn_hex", "0x$mask_hex");
}

sub get_char ($) {
    my $bits = shift;
    if ($bits =~ /([A-Za-z])/) {
	return $1;
    } else {
	return "";
    }
}

sub isolate_args ($$)
{
    my $bits = shift;
    my $bits2 = shift;
    my $code = "";
    my $kset = 0;

    while (1) {
	my $c = get_char ($bits);
	last if ($c eq "");

	my @list = split //, $bits;

	$code .= "    dec->${c}_set = true;\n";
	if ($c eq "k") {
	    $kset = 1;
	}

	my $first = 1;
	for (my $i=0; $i<16; $i++) {
	    # print "$i $c\n";
	    if ($list[$i] eq $c) {
		if ($first) {
		    $code .= "    dec->$c = 0;\n";
		    $first = 0;
		} else {
		    $code .= "    dec->$c <<= 1;\n";
		}
		$code .= "    if ((raw_insn & (1<<".(15-${i}).")) != 0) dec->$c++;\n";
	    }
	}

	($bits =~ s/$c/0/g);
    }

    if (defined ($bits2)) {
	die if ($bits2 ne "kkkkkkkkkkkkkkkk");

	if ($kset) {
	    $code .= "    dec->k <<= 16;\n";
	} else {
	    $code .= "    dec->k = 0;\n";
	    $code .= "    dec->k_set = true;\n";
	}
	$code .= "    {\n";
	$code .= "      unsigned short second;\n";
	$code .= "      second = get_insn_short (addr+2);\n";
	$code .= "      dec->k += second;\n";
	$code .= "      dec->raw_opcode2 = second;\n";
	$code .= "      dec->raw_opcode2_set = true;\n";
	$code .= "      dec->size = 4;\n";
	$code .= "    }\n";
    } else {
	$code .= "    dec->raw_opcode2_set = false;\n";
	$code .= "    dec->size = 2;\n";
    }

    return $code;
}

open INF, "<insns.txt" or die;

open OP_ENUM, ">avr_opcode_enum.ci" or die;
open OP_STR,  ">avr_opcode_to_str.ci" or die;
open OP_SWCH, ">avr_decode_switch.ci" or die;
open OP_MAC,  ">avr_opcode_macros.ci" or die;

my %opcodes;
my %opcode_bitss;
my $opcode_cnt = 0;

while (my $line = <INF>) {
    ($line =~ s/\#(.*)$//);
    next if ($line =~ /^$/);
    
    my @list = split /\s+/, $line;
    my $opcode = "AVR_".$list[0];
    my $print_opcode = $list[0];
    ($print_opcode =~ s/[0-9].*$//);

    # print "opcode: $opcode\n";
    $opcode_cnt++;
    
    die "oops: duplicate opcode $opcode" if (defined ($opcodes{$opcode}));
    $opcodes{$opcode} = $opcode;
    
    print OP_ENUM "  ${opcode},\n";

    print OP_STR "  case ${opcode}:\n";
    print OP_STR "    return \"$print_opcode\";\n\n";

    my $bits = $list[2].$list[3].$list[4].$list[5];
    my $bits2;

    if (defined ($list[6])) {
	$bits2 = $list[6].$list[7].$list[8].$list[9];
    }

    #print "$opcode $bits  ";

    (my $insn_bits, my $insn_mask) = create_masks ($bits);
    #die "oops: $opcode has same bits as $opcode_bitss{$insn_bits}"
    #   if (defined ($opcode_bitss{$insn_bits}));
    $opcode_bitss{$insn_bits} = $opcode;

    my $arg_code = isolate_args ($bits, $bits2);

    print OP_SWCH "  if ((raw_insn & $insn_mask) == $insn_bits) {\n";

    # print OP_SWCH "    if (opcode != OP_NONE) {\n";
    # print OP_SWCH "      printf (\"Hmm: double match for %s and $opcode\\n\",\n";
    # print OP_SWCH "              opcode_to_string (opcode));\n";
    # print OP_SWCH "      exit (-1);\n";
    # print OP_SWCH "    }\n";

    # print OP_SWCH "    printf (\"matched $opcode\\n\");\n";

    print OP_SWCH $arg_code;
    print OP_SWCH "    dec->opcode = $opcode;\n";
    print OP_SWCH "    dec->op_type = $list[1];\n";
    print OP_SWCH "    return;\n";
    print OP_SWCH "  }\n\n";

}

close INF;

print OP_MAC "#define NUM_OPCODES $opcode_cnt\n";

close OP_ENUM;
close OP_STR;
close OP_SWCH;
close OP_MAC;

print "$opcode_cnt opcodes processed.\n";
