// A really simple set implementation and an iterator for it

template<class Key, unsigned MaxSize> class set_iterator;

template<class Key, unsigned MaxSize>
class set {
 public:
  unsigned size;
  Key elts[MaxSize];    // elts[0..size-1] contain data
                        // unique wrt == and sorted using < ordering

  typedef set_iterator<Key,MaxSize> iterator;

  // see comment in state_int_t::init
  void init() {
    size = 0;
  };

  // constructors
  set<Key,MaxSize>() : size(0) {};

#if 0
  // Perfectly good code but it's not used...
  set<Key,MaxSize>(Key a) : size(1) {
    elts[0] = a;
  };
#endif

  void set_to_bottom() {
    size = MaxSize+1;
  };

  bool is_bottom () const {
    return (size > MaxSize);
  };

  bool operator == (const class set<Key,MaxSize> b) const {
    if (this->is_bottom()) {
      return b.is_bottom();
    }
    if (b.is_bottom()) {
      return false;
    }
    if (size != b.size) {
      return false;
    }
    for(unsigned i=0; i<size; ++i) {
      if (elts[i] != b.elts[i]) {
        return false;
      }
    }
    return true;
  };

  bool operator != (const class set<Key,MaxSize> b) const {
    return !(*this == b);
  };

  // returns false if element was already in set
  bool insert(Key a) {
    if (this->is_bottom()) return false;
    // ToDo: use binary search
    // (not that important since we plan to use with MaxSize <= 4)
    unsigned i;
    for(i=0; i<size; ++i) {
      if (a == elts[i]) return false;
      if (a < elts[i]) break;
    }
    // elts[0..i-1] < a < elts[i..size]
    if (size == MaxSize) {
      this->set_to_bottom();
      return true;
    }
    for(unsigned j=size; j != i; --j) {
      elts[j] = elts[j-1];
    }
    elts[i] = a;
    size += 1;
    return true;
  };

  bool remove(Key a) {
    if (this->is_bottom()) return true;
    // ToDo: use binary search
    // (not that important since we plan to use with MaxSize <= 4)
    unsigned i;
    for(i=0; i<size && a < elts[i]; ++i) {
    }
    // elts[0..i-1] < a <= elts[i..size]
    if (i == size || a != elts[i]) {
      return false; // not present
    }
    for(unsigned j=i; j < size-1; ++j) {
      elts[j] = elts[j+1];
    }
    size -= 1;
    return true;  // was present
  };

  bool member(const Key &a) {
    if (this->is_bottom()) return true;
    // ToDo: use binary search
    // (not that important since we plan to use with MaxSize <= 4)
    unsigned i;
    for(i=0; i<size; ++i) {
      if (a == elts[i]) return true;
      if (a < elts[i]) break;
    }
    return false;
  };

  iterator begin() const;
  iterator end()   const;
};

template<class Key, unsigned MaxSize>
ostream& operator << (ostream& os, const class set<Key,MaxSize> b) {
  if (b.is_bottom()) {
    return os << "_|_";
  } else {
    os << "{ ";
    for (unsigned i=0; i<b.size; ++i) {
      os << b.elts[i] << ' ';
    }
    os << '}';
    return os;
  }
};

template<class Key, unsigned MaxSize> 
class set_iterator {
  unsigned x;
  const set<Key,MaxSize> &s;
 public:
  set_iterator<Key,MaxSize>(const set<Key,MaxSize> &_s, unsigned _x) 
    : x(_x), s(_s) 
    {};
  Key operator*() {
    //    assert(0 <= x);
    assert( !s.is_bottom() );
    assert(x < s.size);
    return s.elts[x];
  };
  set_iterator<Key,MaxSize> operator ++ () {
    ++x;
    return *this;
  };
  bool operator == (const class set_iterator<Key,MaxSize> b) const {
    return (x == b.x);
  };
  bool operator != (const class set_iterator<Key,MaxSize> b) const {
    return !(*this == b);
  };
};

template<class Key, unsigned MaxSize> 
inline set_iterator<Key,MaxSize> set<Key,MaxSize>::begin() const {
  assert ( !this->is_bottom());
  return set_iterator<Key,MaxSize>(*this,0);
};

template<class Key, unsigned MaxSize> 
inline set_iterator<Key,MaxSize> set<Key,MaxSize>::end() const {
  assert ( !this->is_bottom());
  return set_iterator<Key,MaxSize>(*this,size);
};

// template<class Key, unsigned MaxSize1, unsigned MaxSize2>
// extern void union2(class set<Key,MaxSize1> &dst, 
//                    const class set<Key,MaxSize2> &src);

// returns false if all elements were already in dst
// (We don't call it union because that is a keyword....)
template<class Key, unsigned MaxSize1, unsigned MaxSize2>
inline
bool union2(class set<Key,MaxSize1> &dst, 
            const class set<Key,MaxSize2> &src) {
  if (dst.is_bottom()) {
    return false;
  } else if (src.is_bottom()) {
    dst.set_to_bottom();
    return true;
  } else {
    bool changed = false;
    for(class set<Key,MaxSize2>::iterator x = src.begin(); x != src.end(); ++x) {
      changed |= dst.insert(*x);
    }
    return changed;
  }
};

