#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>
using std::cout;
using std::endl;
using std::cin;
using std::cerr;
using std::flush;
using std::ostream;
using std::ofstream;
using std::hex;
using std::dec;
using std::ios;
using std::ios_base;

#include <string>
using std::string;

#include <sstream>
using std::ostringstream;

// GNU extension to let me use a C file to create a filebuf
#include <ext/stdio_filebuf.h>
using __gnu_cxx::stdio_filebuf;

#include <string>
using std::string;
using std::max;
using std::min;

#define	die() (__die (__FILE__, __LINE__))
extern void __die (const char *, int) __attribute__ ((noreturn));
#undef assert
#define assert(p) do { if (!(p)) die(); } while (false);

extern void *xmalloc (size_t s);
extern void *xrealloc (void *p, size_t s);

#include "set.h"
#include "vec.h"
#include "queue.h"

/////////////////////////// constants /////////////////////////////////

#define MAX_VECTORS        35
#define MAX_IO_SIZE        64
#define MAX_EXTENDED_IO_SIZE 0xa0
#define MAX_REG_FILE_SIZE  32
#define MAX_STATE_SIZE     (MAX_IO_SIZE + MAX_EXTENDED_IO_SIZE + MAX_REG_FILE_SIZE)
/*
 * believed to be true on all supported AVRs (but definitely not true
 * on small members of the family)
 */
#define PC_SIZE                           2
#define VECTOR_SIZE        ((unsigned int)4)
#define SPL_ADDR      ((io_data_addr_t)0x3d)
#define SPH_ADDR      ((io_data_addr_t)0x3e)
#define SREG_ADDR     ((io_data_addr_t)0x3f)

#define MAX_TASKS    100
/* It seems that 2 should be plenty for MAX_PCS but return instructions
 * are pretty promiscuous.
 */
#define MAX_PCS       100 
#define MAX_RETS      100

/*
 * worst-case stacking of the same interrupt 
 *
 * this is basically a hack that we use to cope with the fact that
 * real kernels sometimes leave all interrupts enabled while in an
 * interrupt handler and this (ostensibly!) isn't a problem in
 * practice
 *
 * TODO: make this configurable on a per-interrupt basis using an
 * input file, and make it a mistake to have a self-stackable
 * interrupt that isn't explicitly bounded
 */
#define MAX_IRQ_RECURSE_DEPTH 1

/*
 * worst-case number of times a cycle in the code with non-zero stack
 * effect can be traversed
 * 
 * This is a hack to cope with the fact that the
 * analysis isn't strong enough (yet!) to see termination of recursion.
 *
 * TODO: make this configurable on a per-function basis using an input
 * file, and make it a mistake to have recursion that isn't explicitly
 * bounded
 */
#define MAX_RECURSE_DEPTH 1

/*
 * initialize SP to point outside of memory on any AVR part
 */
#define INIT_SPH 128
#define INIT_SPL 0
#define INIT_SP ((INIT_SPH<<8)+INIT_SPL)

// used to represent -infinity (top) when computing depths, wcet or
// anything else using the (int,max) domain.
#define NEG_INFINITY (-99)

///////////////////////////// strong types //////////////////////////////

typedef unsigned int reg_t;
#define NO_REG ((reg_t)-1)

typedef unsigned int data_addr_t;
#define NO_ADDR ((data_addr_t)-1)

typedef unsigned int code_addr_t;
#define NO_CODE_ADDR ((code_addr_t)-1)

typedef unsigned int code_ctxt_t;
#define NO_CTX ((code_ctxt_t)-1)

typedef int cost_t;
#define NO_COST ((cost_t)-1)

// An abstract location consists of a concrete address and a context
typedef struct abs_loc {
  code_addr_t addr;
  code_ctxt_t ctx;

  abs_loc() 
    : addr(NO_CODE_ADDR), ctx((code_ctxt_t)0)
  {};
  abs_loc( code_addr_t _addr, code_ctxt_t _ctx)
    : addr(_addr), ctx(_ctx)
  {};
  bool operator == (const struct abs_loc &x) const {
    return(addr==x.addr && ctx==x.ctx);
  }
  bool operator != (const struct abs_loc &x) const {
    return !(*this == x);
  };
  bool operator < (const struct abs_loc &x) const {
    return ((addr < x.addr) 
            || ((addr == x.addr) && (ctx < x.ctx)));
  };
} abs_loc_t;

const abs_loc_t NO_LOC = abs_loc (NO_ADDR, NO_CTX);

struct abs_cmp
{
  bool operator()(const class abs_loc& s1, const class abs_loc& s2) const
  {
    return ((s1.addr < s2.addr) 
            || ((s1.addr == s2.addr) && (s1.ctx < s2.ctx)));
  }
};

typedef unsigned int branch_idx_t;
#define NO_IDX ((branch_idx_t)-1)

class scheduler {
 public:
  code_addr_t addr;
  // int func;
  abs_loc_t stash_ret;
  abs_loc_t stash_head;
};

typedef struct call_node {
  abs_loc_t calling_pc;
  code_addr_t called_addr;
} call_node_t;

typedef unsigned int data_t;
typedef data_t data_val_t;
typedef data_t data_kno_t;

typedef unsigned int io_data_addr_t;

typedef int op_type_t;

typedef int immed_t;

typedef int bit_t;

typedef int disp_t;

typedef int bitmask_t;

typedef data_t status_bit_t;
#define C_BIT ((status_bit_t)0)
#define Z_BIT ((status_bit_t)1)
#define N_BIT ((status_bit_t)2)
#define V_BIT ((status_bit_t)3)
#define S_BIT ((status_bit_t)4)
#define H_BIT ((status_bit_t)5)
#define T_BIT ((status_bit_t)6)
#define I_BIT ((status_bit_t)7)

/////////////////////////// data structures ////////////////////////////

struct int_bit {
  code_addr_t addr;
  bit_t bit;

  bool operator == (const struct int_bit &x) const {
    return ((this->addr == x.addr) && (this->bit == x.bit));
  }
};

const struct int_bit NO_BIT = { NO_CODE_ADDR, -1 };

struct interrupt_vector {
  // interrupt number is implicit -- these will always be in an array
  const char *name;
  code_addr_t start_addr;
  struct int_bit enable, flag;
};

const struct interrupt_vector NO_INT = {
  "",
  NO_CODE_ADDR,
  NO_BIT,
  NO_BIT,
};

struct mem_region {
  data_addr_t offset;
  unsigned int len;
};

struct avr_platform {
  char name[256];
  unsigned int num_vectors;
  const struct interrupt_vector int_vecs[MAX_VECTORS+1];
  unsigned int state_size;
  /*
   * FIXME: in the long run there should probably be a variable-size
   * array of tagged memory regions, and a way to specify different
   * address spaces such as sram, code, eeprom, etc.
   */
  struct mem_region regs;
  struct mem_region io;
  struct mem_region extended_io;
  struct mem_region sram;
};

extern const int NUM_PLATFORMS;
extern const struct avr_platform *all_platforms[];
extern const struct avr_platform *platform;
extern struct interrupt_info int_info[MAX_VECTORS+1];

enum AVR_OPCODE {
  AVR_UNKNOWN = 7701,
  AVR_SET_STACK_PTR,
#include "avr_opcode_enum.ci"
};

enum STACK_OPCODE {
  NO_STACK_OPCODE,
  AVR_DEC_STACK_PTR,
  AVR_INC_STACK_PTR
};

#define SD_UNKNOWN ((int)0xdeafbeef)
#define DEC_UNK_1  ((int)0xdeafbe00)
#define INC_UNK_1  ((int)0xdeafbe01)

/*
 * information about an instruction that's mostly relevant to the
 * decoder 
 */
struct decoder_info {
  unsigned short raw_opcode;
  unsigned short raw_opcode2;
  bool raw_opcode2_set;

  enum AVR_OPCODE opcode;
  op_type_t op_type;
  code_addr_t addr;
  unsigned int size;

  // used to track secondary effects, for example,
  // does this insn increment the stack pointer?
  enum STACK_OPCODE stack_opcode; 
  int stack_delta;
  int stack_opcode_size;

  char *label;

  // raw arguments
  
  int r;
  bool r_set;

  int d;
  bool d_set;

  int K;
  bool K_set;

  int k;
  bool k_set;

  int s;
  bool s_set;

  int A;
  bool A_set;

  int b;
  bool b_set;

  int q;
  bool q_set;

  // processed arguments

  code_addr_t target;
  bool target_set;

  reg_t src_reg;
  bool src_reg_set;

  reg_t dst_reg;
  bool dst_reg_set;

  immed_t immed;
  bool immed_set;

  bit_t bit;
  bool bit_set;

  io_data_addr_t ioaddr;
  bool ioaddr_set;

  disp_t disp;
  bool disp_set;
};

struct branch_elt {
  abs_loc_t target;
  bool dead;
#if WCET_ANALYSIS
  cost_t cost;
#endif

  bool operator == (const struct branch_elt &x) const {
    return (target == x.target
            && dead == x.dead
            );
  };
  bool operator != (const struct branch_elt &x) const {
    return !(*this == x);
  };
};

struct branch_list_t {
  struct branch_elt *list;
  unsigned len;

  // see comment on state_int_t::init
  void init() {
    list = NULL; 
    len = 0;
  };

  branch_list_t() {
    this->init();
  };

  bool operator == (const struct branch_list_t &x) const {
    if (len != x.len) return false;
    for (unsigned i=0; i<len; ++i) {
      if (list[i] != x.list[i]) return false;
    }
    return true;
  };
  bool operator != (const struct branch_list_t &x) const {
    return !(*this == x);
  };
};

typedef struct branch_list_t branch_list;

struct control_info {
  branch_idx_t index_taken;
  branch_idx_t index_not_taken;
  branch_list edges_in;
  branch_list edges_out;
};

struct mem_loc_t {
  data_kno_t known;
  data_val_t val;

  mem_loc_t() : known(0x00), val(0xff) {};
  mem_loc_t(data_val_t v) : known(0xff), val(v) {};
  mem_loc_t(data_val_t v, data_kno_t k) : known(k), val(v) {};
  static const mem_loc_t bottom;
  static const mem_loc_t zero; 

  bool operator == (const struct mem_loc_t &x) const {
    return ((this->known == x.known) && (this->val == x.val));
  };

  bool operator != (const struct mem_loc_t &x) const {
    return !(*this == x);
  };

};
typedef struct mem_loc_t mem_loc;

extern ostream& operator << (ostream& os, const class mem_loc_t &b);

struct call_addr_ctx {
  abs_loc_t call;
  abs_loc_t ret;
  abs_loc_t stashed_ret;
  abs_loc_t stashed_head;
};

extern ostream& operator << (ostream& os, const class abs_loc &b);

class memory {
 public:
  mem_loc core[MAX_STATE_SIZE];  // state of registers
  data_kno_t modified[MAX_STATE_SIZE]; // which bits have been modified

#if TRACK_STACK
  vec<mem_loc> stack;         
  // what we know of local stack frame
  // used to track pushes and pops within this frame
  // to handle code like:
  //    push r6; push r5; call foo; pop r5; pop r6;
#endif

  int stack_top;  // top addr of the current stack
  int shadow_sp;  // copy of the real SP that we use to make sure
                  // we're getting the macro-insns right

  // see comment in state_int_t::init
  void init() {
    for(unsigned i=0; i<MAX_STATE_SIZE; ++i) {
      core[i] = mem_loc(0,0);
      modified[i] = (data_kno_t)0;
    }
    stack_top = INIT_SP;
#ifdef TRACK_STACK
    stack.init();
#endif
  };

  memory() {
    this->init();
  };

  bool operator == (const memory &x) const {
    if (false
#if TRACK_STACK
        || stack != x.stack
#endif
        || stack_top != x.stack_top
        ) return false;
    for(unsigned i=0; i<MAX_STATE_SIZE; ++i) {
      if (core[i] != x.core[i]) return false;
      if (modified[i] != x.modified[i]) return false;
    }
    return true;
  };
  bool operator != (const memory &x) const {
    return !(*this == x);
  };
};

// FIXME: somewhat pointless now since it only has one field? - ADR
struct ret_addr_ctx {
  abs_loc_t ret;

  bool operator == (const struct ret_addr_ctx &x) const {
    return(ret == x.ret);
  };
  bool operator != (const struct ret_addr_ctx &x) const {
    return !(*this == x);
  };
};

struct interrupt_info {
  int max_reenter;
  bool duplicate;
};

/*
 * this struct keeps track of context-sensitive program state at each
 * program point -- a safe estimate of: registers, memory, and i/o,
 * dead branch information, etc.
 */
struct state_inst_t {
  memory mem;         // state after executing operation
  branch_list in_edges;
  branch_list out_edges;

  // abstract interpretation
  int num_rets;
  struct ret_addr_ctx *rets; // for an insn that is the target of a
                             // call, the list of return instructions
  int num_calls;
  struct call_addr_ctx *calls; // for an insn that is the target of a
                               // call, the list of calling addresses

  bool may_return;         // true for target of a call that potentially returns (hack)
  bool is_target;
  bool terminate_analysis;
  int visit_cnt;
  abs_loc_t head;          // entry point of current function

  // stack estimation 
  int wc_stack_depth;
  int path_cnt;            // used to terminate recursion/reentrancy

#if WCET_ANALYSIS
  // List of addresses returned to from here (if a return insn)
  int num_ret_targets;
  abs_loc *ret_targets;
  cost_t wcet;
  // wcet from here to end of function
#endif

  // C++ doesn't let you allocate memory using realloc
  // and then initialize it by invoking the constructor 
  // explicitly.  So, this function is the code that would be
  // in the constructor. 
  // This problem propagates down to types of member which must also 
  // provide an init function in many cases.
  // This is total pish...
  void init() {
    mem.init();
    in_edges.init();
    out_edges.init();
    num_rets = 0;
    rets = NULL;
    num_calls = 0;
    calls = NULL;
    may_return = false;
    is_target = false;
    terminate_analysis = false;
    visit_cnt = 0;
    head = abs_loc();
    wc_stack_depth = 0;
    path_cnt = 0;
#if WCET_ANALYSIS
    num_ret_targets = 0;
    ret_targets     = NULL;
    wcet            = NEG_INFINITY;
#endif
  };

  state_inst_t() {
    this->init();
  };

  bool operator == (const struct state_inst_t &x) const {
#if 0
    // Use this code if you want the old comparision code
    return (0 == memcmp (this,&x,sizeof (struct state_inst_t)));
#else
    // Use this code if you want a more precise comparision
    // Take care to keep it up to date as new fields are added!
    return (   mem        == x.mem
	       && in_edges   == x.in_edges
	       && out_edges  == x.out_edges   
	       && num_rets   == x.num_rets    
	       && rets       == x.rets        
	       && num_calls  == x.num_calls   
	       && calls      == x.calls       
	       && may_return == x.may_return  
	       && is_target  == x.is_target   
	       && terminate_analysis == x.terminate_analysis
	       && visit_cnt      == x.visit_cnt       
	       && head           == x.head            
	       && wc_stack_depth == x.wc_stack_depth  
	       && path_cnt       == x.path_cnt  
#if WCET_ANALYSIS
	       && num_ret_targets  == x.num_ret_targets
	       && ret_targets      == x.ret_targets
	       && wcet             == x.wcet
#endif
	       );
#endif
  };
  bool operator != (const struct state_inst_t &x) const {
    return !(*this == x);
  };

};

typedef struct state_inst_t state_inst;

struct avr_insn_t {
  bool touched;
  bool wc_path;

  struct decoder_info decode;

  struct control_info control;

  // dataflow stuff
  state_inst *states;
  unsigned int state_cnt;
  
  // stack analysis stuff
  int wc_stack_depth;
  int wc_path_depth;

#if WCET_ANALYSIS
  code_addr_t loop_end; // end of this loop if this is start
  bool visited;         // used in various dfs searches
  int dfn;              // depth first number
  int exn_count;        // number of times this instruction executed
#endif
};

typedef struct avr_insn_t avr_insn;

static bool is_end_char(char c) {
  return (c == '\0' || c == '|');
}

// We encode optimization info in the srclocs so we need to
// ignore those fields when comparing srclocs
static bool eq_filename(const char* s1, const char* s2) {
  int i=0; 
  do {
    if (is_end_char(s1[i]) && is_end_char(s2[i])) return true;
    if (s1[i] != s2[i]) return false;
    ++i;      
  } while (1);
}

// Locations in source code
class srcloc {
 public:
  char* filename;
  int line_number;
  
  srcloc(char* f, int l) : filename(f), line_number(l) {};
  srcloc() : filename(0), line_number(-1) {};

  bool operator == (const struct srcloc &x) const {
    return (line_number == x.line_number
            && (filename == x.filename
                || (filename 
                    && x.filename 
                    && eq_filename(filename,x.filename))));
  };

};

extern ostream& operator << (ostream& os, const srcloc &loc);

struct ho_fn_t {
  char* name;   // name of higher order function
  reg_t arg;    // register pair function argument is passed in
};

//////////////////////////// prototypes ///////////////////////////////

extern struct ho_fn_t* lookup_ho_fn (code_addr_t adr);
extern code_addr_t pair_to_addr (data_val_t rv1, data_val_t rv2);

extern bool indirectly_enabled[MAX_VECTORS];

extern void decode_addr (code_addr_t addr);
extern unsigned short get_insn_short (code_addr_t addr);
extern const char* opcode_to_string (enum AVR_OPCODE opcode);
extern bool avr_is_call(enum AVR_OPCODE opcode);
extern bool avr_is_ret(enum AVR_OPCODE opcode);
extern bool avr_is_jmp(enum AVR_OPCODE opcode);
extern cost_t avr_cost(enum AVR_OPCODE opcode);
extern void avr_disasm_insn(code_addr_t addr, FILE* outf);
extern void open_obj_file (const char *fn);
extern char *lookup_text_symbol (code_addr_t addr);
extern void disassemble (const char *fn);

extern long long int count_bits (const memory *s);
extern int count_bits(data_val_t x);

extern void print_verbose_dataflow (code_addr_t addr, FILE *outf, ostream &os);
extern void print_brief_dataflow (code_addr_t addr, FILE *outf);

extern void callgraph_shapes (void);
extern const char *parse_args (const int argc, const char *argv[]);
extern void open_aux_files (const char *input_file);
extern void close_aux_files (void);

extern const char *addr_to_head_label (code_addr_t addr);
extern code_addr_t label_to_addr (const char* label);
extern bool addr_to_srclocs (code_addr_t addr, queue<srcloc> **locs);
extern bool addr_to_srcloc (code_addr_t addr, srcloc *loc);
extern ostream& pp_addr (ostream& os, code_addr_t addr);
extern code_addr_t find_first_srcloc (const srcloc &loc);
extern code_addr_t find_last_srcloc (const srcloc &loc);

extern bool safe_ints_enabled (abs_loc_t cur);
extern bool ints_enabled (abs_loc_t cur);

extern code_addr_t nextPC(code_addr_t PC);
extern abs_loc nextPC(abs_loc PC);

extern code_addr_t prevPC(code_addr_t PC);
extern abs_loc prevPC(abs_loc PC);

extern bool avr_is_call(code_addr_t PC);
extern bool avr_is_call(abs_loc PC);
extern bool avr_is_ret(code_addr_t PC);
extern bool avr_is_ret(abs_loc PC);
extern bool avr_is_jmp(code_addr_t PC);
extern bool avr_is_jmp(abs_loc PC);

extern bool int_known (abs_loc_t cur, int vec);
extern bool int_enabled (abs_loc_t cur, int vec);

extern void control_flow_setup (void);
extern bool control_flow_step (void);
extern bool add_tinyos_task (code_addr_t task_addr);
extern bool is_scheduler (code_addr_t addr);
extern void enqueue_task (code_addr_t addr);

extern bool dataflow_step (void);
extern void report_dead_branches (void);
extern void report_stores (void);
extern void check_dataflow_invariants (void);

extern void enqueue_data (abs_loc_t prev,
			  abs_loc_t n, // previously called 'new'
			  abs_loc_t ret, 
			  abs_loc_t head);
     
extern void stack_scan (void);

extern void add_tinyos_tasks (void);
extern void read_wcet_from_file( const char* fn );
#if WCET_ANALYSIS
extern void wcet_analysis(void);
#endif

extern bool check_inlined_task_posts(void); //NDC

extern void abstract_interp_insn (memory *m,
				  const struct decoder_info *dec);

extern void mark_dead_branches (const state_inst *s,
				const struct decoder_info *dec,
				const struct control_info *c);

extern void add_call_info (abs_loc_t target,
                           abs_loc_t cur,
                           abs_loc_t succ,
                           abs_loc_t ret,
                           abs_loc_t head,
                           branch_idx_t idx);

extern void deferred_add_tinyos_task (call_node_t *);

extern void find_counts (void);

extern void undecode_addr (code_addr_t addr);

/////////////////////////// global variables /////////////////////////

extern queue<char*> icall_funs;

// control over debugging output (set by command line flags)
// the higher the number, the greater the verbosity
extern int debug_data;
extern int debug_depth;
extern int debug_wcet;
extern int debug_control;
extern bool debug_elf;

extern bool do_wcet;

extern int bdd_op_inputs[32];

extern int num_schedulers;
extern class scheduler schedulers[MAX_TASKS];
extern class scheduler* find_scheduler(code_addr_t addr);
extern void add_scheduler(code_addr_t addr, abs_loc_t ret, abs_loc_t head);
extern void check_schedulers(void);

extern int print_max_path;
extern bool verbose;
extern bool print_disasm;
extern const char *disasm_fn;
extern bool quiet;
extern bool print_sum_depth;
extern bool print_whole_depth;
extern bool mark_deepest_path; 
extern unsigned int origin;

extern bool print_ipg;
extern bool print_callgraph;
extern FILE *callgraph_file;
extern FILE *df_cg_file;

extern void initialize_values(const struct avr_platform *p,
                              mem_loc_t m[MAX_STATE_SIZE]);


extern size_t data_seg_size, bss_seg_size;
extern data_addr_t SRAM_TOP, init_stack_ptr;

extern const char *bool_to_str (bool b);

// parsing support
extern bool matches(char * &s, char* prefix);
extern bool skip_spaces0(char * &s);
extern bool skip_spaces(char * &s);
extern bool read_num(char * &s, int *n, int base);
extern bool read_token(char * &s, char **t, int *len);
extern bool read_srcloc(char* &s, srcloc *loc);

extern bool suffix_match(const char *s, const char *pat);
extern char* string_copy(const char* s, int len);

extern queue<char*> srcloc_prefixes;
extern char* drop_srcloc_prefix(char* s);
extern srcloc drop_srcloc_prefix(const srcloc &s);

extern void explode (data_val_t v, int *arr, int offset);
extern data_val_t implode (int *arr, int offset);

extern void set_mem (memory *m, int addr, struct mem_loc_t val);
extern struct mem_loc_t get_mem (memory *m, int addr);

extern void set_cc (memory *m, struct mem_loc_t x, data_t mask);
extern struct mem_loc_t get_cc (memory *m);

extern unsigned int code_size;
extern avr_insn **all_code;
extern bool *sp_consistent;

static inline bool is_valid_state(abs_loc_t loc) {
  return (all_code[loc.addr] && loc.ctx < all_code[loc.addr]->state_cnt);
}

static inline state_inst* state(abs_loc_t loc) {
  assert(is_valid_state(loc));
  return &(all_code[loc.addr]->states[loc.ctx]);
}

#include "bit_helpers.h"

static inline mem_loc_t* memry(abs_loc_t loc, data_addr_t a) {
  return &(state(loc)->mem.core[a]);
}

static inline mem_loc_t* regster(abs_loc_t loc, reg_t r) {
  return memry(loc,reg_to_mem(r));
}

static inline mem_loc_t* ioport(abs_loc_t loc, io_data_addr_t p) {
  return memry(loc,io_to_mem(p));
}

// This is the for handling the calls after a fixed point is reached
extern queue<call_node_t> call_queue;

extern branch_idx_t add_branch (code_addr_t src, 
				code_addr_t dest, 
				bool dead, 
				cost_t cost);
