#include "stacktool.h"

FILE *callgraph_file = NULL;
FILE *df_cg_file = NULL;
int vec_id = 0;

static void print_callgraph_head (FILE *fp, const char *input_file)
{
  fprintf (fp, "// compile like this: ");
  fprintf (fp, "\"dot callgraph.dot -Tps -o callgraph.ps\"\n\n");
  fprintf (fp, "digraph \"%s\" {\n", input_file);
  fprintf (fp, "  orientation=landscape;\n");
  fprintf (fp, "  size=\"8.5,11\";\n");
  fprintf (fp, "  page=\"8.5,11\";\n");
}

void open_aux_files (const char *input_file)
{
  if (print_callgraph) {
    callgraph_file = fopen ("callgraph.dot", "w");
    if (callgraph_file == NULL) die ();
    print_callgraph_head (callgraph_file, input_file);

    df_cg_file = fopen ("callgraph-context.dot", "w");
    if (df_cg_file == NULL) die ();
    print_callgraph_head (df_cg_file, input_file);
  }
}

void close_aux_files (void)
{
  if (callgraph_file != NULL) {
    fprintf (callgraph_file, "}\n");
    (void) fclose (callgraph_file);
  }

  if (df_cg_file != NULL) {
    fprintf (df_cg_file, "}\n");
    (void) fclose (df_cg_file);
  }
}

static void print_color (FILE *file,
			 char *cur_label,
			 bool maybe_enabled,
			 bool def_enabled)
{
  if (cur_label != NULL) {
    if (!quiet && false) printf ("  %s : %s prove disabled,  %s prove enabled\n", 
				 cur_label,
				 (maybe_enabled) ? "CAN'T" : "CAN",
				 (def_enabled) ? "CAN" : "CAN'T");
    
    if (def_enabled) {
      fprintf (file, "  \"%s\" [color = red];\n", cur_label);
    } else if (maybe_enabled) {
      fprintf (file, "  \"%s\" [color = yellow];\n", cur_label);
    } else {
      fprintf (file, "  \"%s\" [color = green];\n", cur_label);
    }
  }
}

static const unsigned int MAX_CTX = 100;

void callgraph_shapes (void)
{
  unsigned int vec;

  for (vec=1; vec < platform->num_vectors; vec++) {
    code_addr_t entry = platform->int_vecs[vec].start_addr;
    code_addr_t addr = all_code[entry]->decode.target;

    fprintf (callgraph_file, "  \"%s\" [shape=box];\n",
	     all_code[addr]->decode.label);

    fprintf (df_cg_file, "  \"%s__0\" [shape=box];\n",
	     all_code[addr]->decode.label);
  }

  fprintf (callgraph_file, "  \"main\" [shape=box];\n");
  fprintf (df_cg_file, "  \"main__0\" [shape=box];\n");

  fprintf (callgraph_file, "  { rank = same; \"main\"; ");
  fprintf (df_cg_file, "  { rank = same; \"main__0\"; ");

  for (vec=1; vec < platform->num_vectors; vec++) {
    code_addr_t entry = platform->int_vecs[vec].start_addr;
    code_addr_t addr = all_code[entry]->decode.target;

    fprintf (callgraph_file, "\"%s\"; ",
	     all_code[addr]->decode.label);

    fprintf (df_cg_file, "\"%s__0\"; ",
	     all_code[addr]->decode.label);
  }

  fprintf (callgraph_file, " };\n");
  fprintf (df_cg_file, " };\n");

  {
    char *cur_label = NULL;
    bool maybe_enabled = false, def_enabled = false;
    bool maybe_enabled_ctx[MAX_CTX], def_enabled_ctx[MAX_CTX];
    unsigned int j;
    code_ctxt_t ctx;
    code_addr_t idx = NO_CODE_ADDR;
    code_addr_t addr;

    if (!quiet && false) {
      printf ("list of functions by their interrupt characteristics:\n");
    }

    for (ctx = (code_ctxt_t)0; ctx < (code_ctxt_t)MAX_CTX; ctx++) {
      maybe_enabled_ctx[ctx] = false;
      def_enabled_ctx[ctx] = false;
    }

    for (addr=0; addr < (code_addr_t) code_size; addr++) {

      if (all_code[addr] == NULL) continue;

      if (all_code[addr]->decode.label != NULL) {
	if (cur_label != NULL) {
	  print_color (callgraph_file, cur_label, maybe_enabled, def_enabled);
	  assert (idx != NO_CODE_ADDR);
	  for (j=0; j<all_code[idx]->state_cnt; j++) {
	    char buf[1000];
	    (void) snprintf (buf, 1000, "%s__%d", cur_label, j);
	    print_color (df_cg_file, buf, maybe_enabled_ctx[j], def_enabled_ctx[j]);
	  }
	}
	  
	cur_label = all_code[addr]->decode.label;
	idx = addr;
	maybe_enabled = false;
	def_enabled = false;
	for (j=0; j<MAX_CTX; j++) {
	  maybe_enabled_ctx[j] = false;
	  def_enabled_ctx[j] = false;
	}
      }

      for (j=0; j<all_code[addr]->state_cnt; j++) {
	abs_loc_t l = abs_loc (addr, j);
	if (safe_ints_enabled (l)) {
	  maybe_enabled = true;
	  if (j < MAX_CTX) maybe_enabled_ctx[j] = true;
	}
	if (ints_enabled (l)) {
	  def_enabled = true;
	  if (j < MAX_CTX) def_enabled_ctx[j] = true;
	}
      }
    }
    print_color (callgraph_file, cur_label, maybe_enabled, def_enabled);
    assert (idx != NO_CODE_ADDR);
    for (j=0; j<all_code[idx]->state_cnt; j++) {
      char buf[1000];
      (void) snprintf (buf, 1000, "%s__%d", cur_label, j);
      print_color (df_cg_file, buf, maybe_enabled_ctx[j], def_enabled_ctx[j]);
    }
  }
}

void __die (const char *file, int line)
{
  cout << flush;
  cerr << flush;
  fflush (stdout);
  fflush (stderr);
  fprintf (stderr, "FATAL ERROR: at line %d of file %s\n", line, file);
  fflush (stderr);
  exit (EXIT_FAILURE);
}

static code_addr_t addr_to_head (code_addr_t addr)
{
  if (addr < (code_addr_t) (VECTOR_SIZE * platform->num_vectors)) {
    return addr;
  }
  assert( addr < (code_addr_t) code_size );

  for (code_addr_t i=addr; i >= (code_addr_t) (VECTOR_SIZE * platform->num_vectors); i -= 1) {
    if (all_code[i] != NULL &&
	all_code[i]->decode.label != NULL) {
      return i;
    }
  }

  return NO_CODE_ADDR;
}

static const char *entry_labels[] = {
  "[vec00]",  "[vec01]",  "[vec02]",  "[vec03]",
  "[vec04]",  "[vec05]",  "[vec06]",  "[vec07]",
  "[vec08]",  "[vec09]",  "[vec10]",  "[vec11]",
  "[vec12]",  "[vec13]",  "[vec14]",  "[vec15]",
  "[vec16]",  "[vec17]",  "[vec18]",  "[vec19]",
  "[vec20]",  "[vec21]",  "[vec22]",  "[vec23]",
  "[vec24]",  "[vec25]",  "[vec26]",  "[vec27]",
  "[vec28]",  "[vec29]",  "[vec30]",  "[vec31]",
  "[vec32]",  "[vec33]",  "[vec34]",  "[vec35]"
};

const char *addr_to_head_label (code_addr_t addr)
{
  assert(platform->num_vectors <= (sizeof(entry_labels)/sizeof(const char*)));
  if (addr == NO_CODE_ADDR) return "--";

  if (addr < (code_addr_t) (VECTOR_SIZE * platform->num_vectors)) {
    assert (addr % VECTOR_SIZE == 0);
    unsigned int vec = addr / VECTOR_SIZE;
    assert (vec < platform->num_vectors);
    return entry_labels[vec];
  }

  assert( addr < (code_addr_t) code_size );
  if (all_code[addr] == NULL) die ();

  code_addr_t head = addr_to_head (addr);
  if (head == NO_CODE_ADDR) {
    return "--";
  } else {
    assert( head < (code_addr_t) code_size );
    assert( all_code[head] != NULL );
    assert( all_code[head]->decode.label != NULL );
    return all_code[head]->decode.label;
  }
}

code_addr_t nextPC(code_addr_t PC) {
  return (PC + (code_addr_t) all_code[PC]->decode.size);
}

abs_loc nextPC(abs_loc PC) {
  return abs_loc(nextPC(PC.addr),PC.ctx);
}

bool avr_is_call(code_addr_t PC) {
  return all_code[PC] && avr_is_call(all_code[PC]->decode.opcode);
}

bool avr_is_call(abs_loc PC) {
  return avr_is_call(PC.addr);
}

bool avr_is_ret(code_addr_t PC) {
  return all_code[PC] && avr_is_ret(all_code[PC]->decode.opcode);
}

bool avr_is_ret(abs_loc PC) {
  return avr_is_ret(PC.addr);
}

bool avr_is_jmp(code_addr_t PC) {
  return all_code[PC] && avr_is_jmp(all_code[PC]->decode.opcode);
}

bool avr_is_jmp(abs_loc PC) {
  return avr_is_jmp(PC.addr);
}

code_addr_t prevPC(code_addr_t PC) {
  code_addr_t prev = PC;
  // Hack: search backwards for previous instruction
  do {
    --prev;
  } while (all_code[prev]==NULL || nextPC(prev)!=PC);
  assert((int)prev >= 0);
  return prev;
}

abs_loc prevPC(abs_loc PC) {
  return abs_loc(prevPC(PC.addr),PC.ctx);
}

const char *bool_to_str (bool b)
{
  return b ? "true" : "false";
}

void *xmalloc (size_t s)
{
  void *p = malloc (s);
  if (p==NULL) die();
  return p;
}

void *xrealloc (void *p, size_t s)
{
  void *p2 = realloc (p, s);
  if (p2==NULL) die();
  return p2;
}

////////////////////////////////////////////////////////////////
// Parsing support
//
// Functions return bool indicating success and increment char* argument
// to indicate progress.  Typical usage (equivalent to /^foo\s+(\d+)\s*$/):
//
//      matches(s,(char*)"foo") 
//      && skip_spaces(s) && read_num(s,&m,10)
//      && skip_spaces0(s) && *s == '\0'
//
////////////////////////////////////////////////////////////////

bool matches(char * &s, char* prefix) {
  int l = strlen(prefix);
  if (0 == strncmp(s,prefix,l)) {
    s += l;
    return true;
  } else {
    return false;
  }
}

bool read_token(char * &s, char **t, int *len) {
  *t = s;
  while (!isspace(*s)) {
    ++s;
  }
  *len = s - *t; // length
  return (s != *t);
}

// skip 0 or more spaces
bool skip_spaces0(char * &s) {
  while (isspace(*s)) {
    ++s;
  }
  return true;
}

// skip 1 or more spaces
bool skip_spaces(char * &s) {
  if (isspace(*s)) {
    skip_spaces0(s);
    return true;
  } else {
    return false;
  }
}

bool read_num(char * &s, int *n, int base) {
  char* s2;
  int r = strtol(s,&s2,base);
  if (s == s2) {
    return false;
  } else {
    s = s2;
    *n = r;
    return true;
  }
}

// A srcloc looks like this:
//
//   D:/cygwin/opt/tinyos-1.x/tos/system/sched.c::108
//   ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^        filename
//                                              ^^      separator
//                                                ^^^   linenumber
//
// fn is allocated using malloc and should be freed by caller
bool read_srcloc(char* &s, srcloc *loc) {
  char* t = s;

  char* start = t;
  while (!isspace(*t) && !(*t==':' && *(t+1)==':')) {
    ++t;
  }
  char* end = t;
  if (end == start) return false;

  if (!matches(t,(char*)"::")) return false;

  int ln;
  if (!read_num(t,&ln,10)) return false;

  loc->filename = (char*)xmalloc(1 + end-start);
  strncpy(loc->filename,start,end-start);
  loc->filename[end-start] = 0;
  loc->line_number = ln;
  s=t;
  return true;
}

////////////////////////////////////////////////////////////////
// Other string handling code
////////////////////////////////////////////////////////////////

bool suffix_match(const char *s, const char *pat) {
  int m = strlen(s);
  int n = strlen(pat);
  return (strcmp(s+m-n,pat) == 0);
}

char* string_copy(const char* s, int n) {
  char* r = (char*)xmalloc(n+1);
  strncpy(r,s,n);
  r[n]=0;
  return r;
}

queue<char*> srcloc_prefixes;

char* drop_srcloc_prefix(char* s) {
  for(queue<char*>::iterator j = srcloc_prefixes.begin(); 
      j != srcloc_prefixes.end(); 
      ++j) {
    int i;
    for(i=0; (*j)[i] == s[i] && s[i]; ++i) {
    }
    if ((*j)[i] == 0 || isspace((*j)[i])) {
      // cout << "Left with suffix " << s+i << endl;
      return (s+i);
    }
    // cout << (*j)[i] << "Rejecting prefix " << (*j) << endl;
  }
  // cout << "Couldn't find matching prefix for " << s << endl;
  return s;
}

srcloc drop_srcloc_prefix(const srcloc &s) {
  return srcloc(drop_srcloc_prefix(s.filename), s.line_number);
}
  
////////////////////////////////////////////////////////////////
// End
////////////////////////////////////////////////////////////////

